/// SPDX-License-Identifier: GPL-3.0-or-later
/// SPDX-FileCopyrightText: Copyright © 2016-2021 Tony Garnock-Jones <tonyg@leastfixedpoint.com>

import {
    isToken, isTokenType, replace, commaJoin, startPos, fixPos, joinItems,
    laxRead, itemText,

    Items, Pattern, Templates, Substitution, TokenType,
    SourceMap, CodeWriter, TemplateFunction, Token, SpanIndex,
} from '../syntax/index.js';
import {
    SyndicateParser, SyndicateTypedParser,
    Identifier,
    TurnAction,
    Statement,
    Binder,

    compilePattern,
} from './grammar.js';

export function stripShebang(items: Items): Items {
    if ((items.length > 0) &&
        isToken(items[0]) &&
        items[0].text.startsWith('#!')) {
        while (items.length > 0 && !isTokenType(items[0], TokenType.NEWLINE)) items.shift();
    }
    return items;
}

export type ModuleType ='es6' | 'require' | 'global';

export interface CompileOptions {
    source: string,
    name?: string,
    runtime?: string,
    module?: ModuleType,
    global?: string,
    typescript?: boolean,
}

export interface CompilerOutput {
    text: string,
    map: SourceMap,
    targetToSourceMap: SpanIndex<Token>;
    sourceToTargetMap: SpanIndex<number>;
}

function receiverFor(s: TurnAction): Substitution {
    return (s.implicitTurn) ? 'thisTurn.' : '.';
}

export class ExpansionContext {
    readonly parser: SyndicateParser;
    readonly moduleType: ModuleType;
    hasBootableBootProc: boolean = false;
    readonly typescript: boolean;
    nextIdNumber = 0;

    constructor(moduleType: ModuleType,
                typescript: boolean)
    {
        this.parser = typescript ? new SyndicateTypedParser : new SyndicateParser();
        this.moduleType = moduleType;
        this.typescript = typescript;
    }

    quasiRandomId(): string {
        return '__SYNDICATE__id_' + (this.nextIdNumber++);
    }

    argDecl(t: TemplateFunction, name: Substitution, type: Substitution): Items {
        return (this.typescript) ? t`${name}: ${type}` : t`name`;
    }

    turnDecl(t: TemplateFunction): Items {
        return this.argDecl(t, 'thisTurn', '__SYNDICATE__.Turn');
    }
}

function stringifyId(i: Identifier): Items {
    return [ { ... i, type: TokenType.STRING, text: JSON.stringify(i.text) } ];
}

function binderTypeGuard(t: TemplateFunction): (binder: Binder, index: number) => Items {
    return (binder, index) => {
        if (binder.id.text[0] === '_') {
            return t`${`/* Ignoring underscore-prefixed binder ${binder.id.text} */`}`;
        }

        const raw = t`__vs[${''+index}]`;
        const bind = t`const ${[binder.id]} = ${raw};`;
        if (binder.type === void 0) {
            return bind;
        } else {
            const typeText = itemText(binder.type);
            switch (typeText) {
                case 'boolean':
                case 'string':
                case 'number':
                case 'symbol':
                    return t`if (typeof (${raw}) !== ${JSON.stringify(typeText)}) return;\n${bind}`;
                case 'any':
                    return bind;
                default:
                    // TODO: something better than this!!
                    throw new Error(`Unhandled binding type: ${JSON.stringify(typeText)}`);
            }
        }
    };
}

export function expand(tree: Items, ctx: ExpansionContext): Items {
    const macro = new Templates(undefined, { extraDelimiters: ':' });

    function terminalWrap(t: TemplateFunction, isTerminal: boolean, body: Statement): Statement {
        if (isTerminal) {
            return t`thisTurn._stop(thisTurn.activeFacet, (${ctx.turnDecl(t)}) => {${body}})`
        } else {
            return body;
        }
    }

    function x<T>(p: Pattern<T>, f: (v: T, t: TemplateFunction) => Items) {
        tree = replace(tree, p, (v, start) => f(v, macro.template(fixPos(start))));
    }

    function xf<T extends TurnAction>(p: Pattern<T>, f: (v: T, t: TemplateFunction) => Items) {
        x(p, (v, t) => t`${receiverFor(v)}${f(v, t)}`);
    }

    const walk = (tree: Items): Items => expand(tree, ctx);
    const maybeWalk = (tree?: Items) : Items | undefined => (tree === void 0) ? tree : walk(tree);

    xf(ctx.parser.duringStatement, (s, t) => {
        // TODO: untyped template
        const sa = compilePattern(s.pattern);
        return t`assertDataflow((${ctx.turnDecl(t)}) => ({
            target: currentSyndicateTarget,
            assertion: __SYNDICATE__.fromObserve(__SYNDICATE__.Observe({
                pattern: ${sa.skeleton},
                observer: thisTurn.ref(__SYNDICATE__.assertionFacetObserver(
                    (${ctx.turnDecl(t)}, ${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}) => {
                        if (Array.isArray(__vs)) {
${joinItems(sa.captureBinders.map(binderTypeGuard(t)), '\n')}
                            ${walk(s.body)}
                        }
                    }
                ))
            })),
        }));`;
    });

    xf(ctx.parser.spawn, (s, t) => {
        // TODO: parentBinders, parentInits
        let body = walk(s.body);
        /*
        let assertions = (s.initialAssertions.length > 0)
            ? t`, new __SYNDICATE__.Set([${commaJoin(s.initialAssertions.map(walk))}])`
            : ``;
        */
        const n = s.name === void 0 ? '' : t` thisTurn.activeFacet.actor.name = ${walk(s.name)};`;
        return t`_spawn${s.isLink ? 'Link': ''}((${ctx.turnDecl(t)}) => {${n} ${body} });`;
    });

    x(ctx.parser.fieldDeclarationStatement, (s, t) => {
        const ft = ctx.typescript ? t`<${s.field.type ?? '__SYNDICATE__.AnyValue'}>` : '';
        return t`const ${[s.field.id]} = thisTurn.field${ft}(${maybeWalk(s.init) ?? 'void 0'}, ${stringifyId(s.field.id)});`;
    });

    x(ctx.parser.atStatement, (s, t) => {
        return t`(((${ctx.argDecl(t, 'currentSyndicateTarget', '__SYNDICATE__.Ref')}) => {${walk(s.body)}})(${walk(s.target)}));`;
    });

    x(ctx.parser.createExpression, (s, t) => {
        return t`thisTurn.ref(${walk(s.entity)})`;
    });

    xf(ctx.parser.assertionEndpointStatement, (s, t) => {
        if (s.isDynamic) {
            if (s.test == void 0) {
                return t`assertDataflow((${ctx.turnDecl(t)}) => ({ target: currentSyndicateTarget, assertion: ${walk(s.template)} }));`;
            } else {
                return t`assertDataflow((${ctx.turnDecl(t)}) => (${walk(s.test)})
                       ? ({ target: currentSyndicateTarget, assertion: ${walk(s.template)} })
                       : ({ target: void 0, assertion: void 0 }));`;
            }
        } else {
            if (s.test == void 0) {
                return t`assert(currentSyndicateTarget, ${walk(s.template)});`;
            } else {
                return t`replace(currentSyndicateTarget, void 0, (${walk(s.test)}) ? (${walk(s.template)}) : void 0);`;
            }
        }
    });

    xf(ctx.parser.dataflowStatement, (s, t) =>
        t`_dataflow((${ctx.turnDecl(t)}) => {${walk(s.body)}});`);

    xf(ctx.parser.eventHandlerEndpointStatement, (s, t) => {
        if (s.triggerType === 'dataflow') {
            return t`withSelfDo((${ctx.turnDecl(t)}) => { dataflow { if (${walk(s.predicate)}) { ${terminalWrap(t, s.terminal, walk(s.body))} } } });`;
        }

        if (s.triggerType === 'stop') {
            return t`activeFacet.onStop((${ctx.turnDecl(t)}) => {${walk(s.body)}});`;
        }

        const sa = compilePattern(s.pattern);
        const guardBody = (body: Statement) => t`if (Array.isArray(__vs)) {
${joinItems(sa.captureBinders.map(binderTypeGuard(t)), '\n')}
            ${body}
        }`;

        let entity: Items;
        switch (s.triggerType) {
            case 'asserted':
                entity = t`{
                    assert(${ctx.turnDecl(t)}, ${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}, __handle: __SYNDICATE__.Handle) {
                        ${guardBody(terminalWrap(t, s.terminal, walk(s.body)))}
                    }
                }`;
                break;
            case 'retracted':
                entity = t`__SYNDICATE__.assertionObserver((${ctx.turnDecl(t)}, ${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}) => {
                    ${guardBody(t`return (${ctx.turnDecl(t)}) => { ${terminalWrap(t, s.terminal, walk(s.body))} };`)}
                })`;
                break;
            case 'message':
                entity = t`{
                    message(${ctx.turnDecl(t)}, ${ctx.argDecl(t, '__vs', '__SYNDICATE__.AnyValue')}) {
                        ${guardBody(terminalWrap(t, s.terminal, walk(s.body)))}
                    }
                }`;
                break;
        }

        const assertion = t`__SYNDICATE__.fromObserve(__SYNDICATE__.Observe({
            pattern: ${sa.skeleton},
            observer: thisTurn.ref(${entity}),
        }))`;

        if (s.isDynamic) {
            return t`assertDataflow((${ctx.turnDecl(t)}) => ({
                target: currentSyndicateTarget,
                assertion: ${assertion},
            }));`;
        } else {
            return t`replace(currentSyndicateTarget, void 0, ${assertion});`;
        }
    });

    x(ctx.parser.typeDefinitionStatement, (s, t) => {
        const l = `Symbol.for(${JSON.stringify(s.label.text)})`;
        const fns = JSON.stringify(s.fields.map(f => f.id.text));
        const formatBinder = (b: Binder) => t`${[b.id]}: ${b.type ?? '__SYNDICATE__.AnyValue'}`;
        const fs = ctx.typescript
            ? t`<{${commaJoin(s.fields.map(formatBinder))}}, __SYNDICATE__.Ref>`
            : '';
        return t`const ${[s.label]} = __SYNDICATE__.Record.makeConstructor${fs}()(${maybeWalk(s.wireName) ?? l}, ${fns});`;
    });

    xf(ctx.parser.messageSendStatement, (s, t) => t`message(currentSyndicateTarget, ${walk(s.expr)});`);

    xf(ctx.parser.reactStatement, (s, t) => {
        return t`facet((${ctx.turnDecl(t)}) => {${s.body}});`;
    });

    x(ctx.parser.bootStatement, (s, t) => {
        ctx.hasBootableBootProc = s.formals.length == 0;
        const body = t`{${walk(s.body)}}`;
        const bootFormals = commaJoin([ctx.turnDecl(t), ... s.formals.map(
            b => ctx.argDecl(t, [b.id], b.type ?? '__SYNDICATE__.AnyValue'))]);
        switch (ctx.moduleType) {
            case 'es6':
                return t`export function boot(${bootFormals}) ${body}`;
            case 'require':
                return t`module.exports.boot = (${bootFormals}) => ${body};`;
            case 'global':
                return t`function boot(${bootFormals}) ${body}`;
        }
    });

    xf(ctx.parser.stopStatement, (s, t) =>
        t`withSelfDo((${ctx.turnDecl(t)}) => thisTurn._stop(thisTurn.activeFacet, (${ctx.turnDecl(t)}) => {${walk(s.body)}});`)

    return tree;
}

export function compile(options: CompileOptions): CompilerOutput {
    const inputFilename = options.name ?? '/dev/stdin';

    // console.info(`Syndicate: compiling ${inputFilename}`);

    const source = options.source;
    const moduleType = options.module ?? 'es6';
    const typescript = options.typescript ?? false;

    const start = startPos(inputFilename);
    let tree = stripShebang(laxRead(source, { start, extraDelimiters: ':' }));
    const end = tree.length > 0 ? tree[tree.length - 1].end : start;

    const macro = new Templates(undefined, { extraDelimiters: ':' });

    const ctx = new ExpansionContext(moduleType, typescript);

    tree = expand(tree, ctx);

    const ts = macro.template(fixPos(start));
    const te = macro.template(fixPos(end));

    if (ctx.hasBootableBootProc) {
        let bp;
        switch (moduleType) {
            case 'es6':
            case 'global':
                bp = te`boot`;
                break;
            case 'require':
                bp = te`module.exports.boot`;
                break;
        }
        tree = te`${tree}\nif (typeof module !== 'undefined' && ((typeof require === 'undefined' ? {main: void 0} : require).main === module)) __SYNDICATE__.Actor.boot(${bp});`;
    }

    {
        const runtime = options.runtime ?? '@syndicate-lang/core';
        switch (moduleType) {
            case 'es6':
                tree = ts`import * as __SYNDICATE__ from ${JSON.stringify(runtime)};\n${tree}`;
                break;
            case 'require':
                tree = ts`const __SYNDICATE__ = require(${JSON.stringify(runtime)});\n${tree}`;
                break;
            case 'global':
                tree = ts`const __SYNDICATE__ = ${runtime};\n${tree}`;
                break;
        }
    }

    const cw = new CodeWriter(inputFilename);
    cw.emit(tree);

    const text = cw.text;

    return {
        text,
        map: cw.map,
        targetToSourceMap: cw.targetToSourceMap.index(),
        sourceToTargetMap: cw.sourceToTargetMap.index(),
    };
}
